//
//  ChecksumTest.cpp
//
//  Created  by Setsu on 7/11/2010.
//  Modified by Setsu on May 30, 2013.
//  Copyright 2010-2013 RoundSquare Inc. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person or organization
// obtaining a copy of the software and accompanying documentation covered by
// this license (the "Software") to use, reproduce, display, distribute,
// execute, and transmit the Software, and to prepare derivative works of the
// Software, and to permit third-parties to whom the Software is furnished to
// do so, all subject to the following:
// 
// The copyright notices in the Software and this entire statement, including
// the above license grant, this restriction and the following disclaimer,
// must be included in all copies of the Software, in whole or in part, and
// all derivative works of the Software, unless such copies or derivative
// works are solely in the form of machine-executable object code generated by
// a source language processor.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
// FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//

//----------------------------------------
//	include
//----------------------------------------
#include <Poco/Buffer.h>
#include <Poco/Random.h>
#include <Poco/Stopwatch.h>
#include <Poco/Format.h>
#include <Poco/Checksum.h>

#include <string>

#include "ScopedLogMessage.h"
#include "PrepareConsoleLogger.h"

//----------------------------------------
//	const
//----------------------------------------
const std::size_t kBufferSize = 64*1024*1024;
const int kNumTrial = 4;

//----------------------------------------
//	AdjustAllocationBytes
//		adjust number of allocation bytes for low memory machine
//----------------------------------------
#if defined(POCO_OS_FAMILY_WINDOWS)
#include <new.h>

int ThrowBadAlloc(std::size_t)
{
	throw std::bad_alloc();
	return 0;
}
#endif

std::size_t AdjustAllocationBytes(ScopedLogMessage& msg, std::size_t bytesWanted)
{
	std::size_t bufferSize = bytesWanted;

	bool memoryAvailable = false;
	char* tempPtr = NULL;

#if defined(POCO_OS_FAMILY_WINDOWS)
	_set_new_handler(ThrowBadAlloc);
#endif

	while(!memoryAvailable)
	{
		try
		{
			tempPtr = new char [bufferSize];
			memoryAvailable = true;
		}
		catch(...)
		{
			bufferSize /= 2;
		}
		delete [] tempPtr;
		tempPtr = NULL;
	}
	msg.Message(Poco::format(" Number of buffer allocation bytes: %z", bufferSize));
	return bufferSize;
}

//----------------------------------------
//	main
//----------------------------------------
int main(int /*argc*/, char** /*argv*/)
{
	PrepareConsoleLogger logger(Poco::Logger::ROOT, Poco::Message::PRIO_INFORMATION);

	ScopedLogMessage msg("ChecksumTest ", "start", "end");

	Poco::Checksum checksum[] = {	Poco::Checksum(Poco::Checksum::TYPE_ADLER32)
								,	Poco::Checksum(Poco::Checksum::TYPE_CRC32)	};
	const int kNumChecksumTypes = sizeof(checksum)/sizeof(checksum[0]);

	double time[kNumChecksumTypes][kNumTrial];
	Poco::UInt32 sum[kNumChecksumTypes][kNumTrial];

	Poco::Stopwatch stopwatch;
	Poco::Random random;
	Poco::Buffer<char> buffer(AdjustAllocationBytes(msg, kBufferSize));

	for(int trial=0; trial<kNumTrial; ++trial)
	{
		std::fill(buffer.begin(), buffer.end(), random.nextChar());

		for(int type=0; type<kNumChecksumTypes; ++type)
		{
			stopwatch.restart();
			checksum[type].update(buffer.begin(), static_cast<unsigned>(buffer.size()));
			stopwatch.stop();
			time[type][trial] = (1000.0 * stopwatch.elapsed()) / stopwatch.resolution();
			sum[type][trial] = checksum[type].checksum();
		}
	}

	msg.Message("         TYPE_ADLER32                 TYPE_CRC32");
	msg.Message("      checksum     time           checksum     time");
	msg.Message(" =============================+=======================");
	for(int trial=0; trial<kNumTrial; ++trial)
	{
		msg.Message(Poco::format(" #%d: 0x%08X  %7.3fmSec     0x%08X  %7.3fmSec"
								, trial+1
								,  sum[Poco::Checksum::TYPE_ADLER32][trial]
								, time[Poco::Checksum::TYPE_ADLER32][trial]
								,  sum[Poco::Checksum::TYPE_CRC32][trial]
								, time[Poco::Checksum::TYPE_CRC32][trial]));
	}

	return 0;
}
